@@ -3,12 +3,12 @@
  * pg_operator.c
  *	  routines to support manipulation of the pg_operator relation
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2006, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/catalog/pg_operator.c,v 1.84 2003/09/25 06:57:58 petere Exp $
+ *	  $PostgreSQL: pgsql/src/backend/catalog/pg_operator.c,v 1.98 2006/07/14 14:52:18 momjian Exp $
  *
  * NOTES
  *	  these routines moved here from commands/define.c and somewhat cleaned up.
@@ -18,11 +18,12 @@
 #include "postgres.h"
 
 #include "access/heapam.h"
-#include "catalog/catname.h"
+#include "access/xact.h"
 #include "catalog/dependency.h"
 #include "catalog/indexing.h"
-#include "catalog/namespace.h"
+#include "catalog/pg_namespace.h"
 #include "catalog/pg_operator.h"
+#include "catalog/pg_proc.h"
 #include "catalog/pg_type.h"
 #include "miscadmin.h"
 #include "parser/parse_func.h"
@@ -57,7 +58,7 @@ static Oid get_other_operator(List *otherOp,
 				   Oid leftTypeId, Oid rightTypeId,
 				   bool isCommutator);
 
-static void makeOperatorDependencies(HeapTuple tuple, Oid pg_operator_relid);
+static void makeOperatorDependencies(HeapTuple tuple);
 
 
 /*
@@ -89,10 +90,10 @@ validOperatorName(const char *name)
 
 	/*
 	 * For SQL92 compatibility, '+' and '-' cannot be the last char of a
-	 * multi-char operator unless the operator contains chars that are not
-	 * in SQL92 operators. The idea is to lex '=-' as two operators, but
-	 * not to forbid operator names like '?-' that could not be sequences
-	 * of SQL92 operators.
+	 * multi-char operator unless the operator contains chars that are not in
+	 * SQL92 operators. The idea is to lex '=-' as two operators, but not to
+	 * forbid operator names like '?-' that could not be sequences of SQL92
+	 * operators.
 	 */
 	if (len > 1 &&
 		(name[len - 1] == '+' ||
@@ -174,8 +175,9 @@ OperatorLookup(List *operatorName,
 	Oid			operatorObjectId;
 	RegProcedure oprcode;
 
-	operatorObjectId = LookupOperName(operatorName, leftObjectId,
-									  rightObjectId, true);
+	operatorObjectId = LookupOperName(NULL, operatorName,
+									  leftObjectId, rightObjectId,
+									  true, -1);
 	if (!OidIsValid(operatorObjectId))
 	{
 		*defined = false;
@@ -227,14 +229,14 @@ OperatorShellMake(const char *operatorName,
 	}
 
 	/*
-	 * initialize values[] with the operator name and input data types.
-	 * Note that oprcode is set to InvalidOid, indicating it's a shell.
+	 * initialize values[] with the operator name and input data types. Note
+	 * that oprcode is set to InvalidOid, indicating it's a shell.
 	 */
 	i = 0;
 	namestrcpy(&oname, operatorName);
 	values[i++] = NameGetDatum(&oname); /* oprname */
 	values[i++] = ObjectIdGetDatum(operatorNamespace);	/* oprnamespace */
-	values[i++] = Int32GetDatum(GetUserId());	/* oprowner */
+	values[i++] = ObjectIdGetDatum(GetUserId());		/* oprowner */
 	values[i++] = CharGetDatum(leftTypeId ? (rightTypeId ? 'b' : 'r') : 'l');	/* oprkind */
 	values[i++] = BoolGetDatum(false);	/* oprcanhash */
 	values[i++] = ObjectIdGetDatum(leftTypeId); /* oprleft */
@@ -253,7 +255,7 @@ OperatorShellMake(const char *operatorName,
 	/*
 	 * open pg_operator
 	 */
-	pg_operator_desc = heap_openr(OperatorRelationName, RowExclusiveLock);
+	pg_operator_desc = heap_open(OperatorRelationId, RowExclusiveLock);
 	tupDesc = pg_operator_desc->rd_att;
 
 	/*
@@ -269,7 +271,7 @@ OperatorShellMake(const char *operatorName,
 	CatalogUpdateIndexes(pg_operator_desc, tup);
 
 	/* Add dependencies for the entry */
-	makeOperatorDependencies(tup, RelationGetRelid(pg_operator_desc));
+	makeOperatorDependencies(tup);
 
 	heap_freetuple(tup);
 
@@ -391,7 +393,7 @@ OperatorCreate(const char *operatorName,
 				restOid,
 				joinOid;
 	bool		selfCommutator = false;
-	Oid			typeId[FUNC_MAX_ARGS];
+	Oid			typeId[4];		/* only need up to 4 args here */
 	int			nargs;
 	NameData	oname;
 	TupleDesc	tupDesc;
@@ -409,7 +411,7 @@ OperatorCreate(const char *operatorName,
 	if (!OidIsValid(leftTypeId) && !OidIsValid(rightTypeId))
 		ereport(ERROR,
 				(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
-		errmsg("at least one of leftarg or rightarg must be specified")));
+		   errmsg("at least one of leftarg or rightarg must be specified")));
 
 	if (!(OidIsValid(leftTypeId) && OidIsValid(rightTypeId)))
 	{
@@ -417,11 +419,11 @@ OperatorCreate(const char *operatorName,
 		if (commutatorName)
 			ereport(ERROR,
 					(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
-				  errmsg("only binary operators can have commutators")));
+					 errmsg("only binary operators can have commutators")));
 		if (joinName)
 			ereport(ERROR,
 					(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
-			 errmsg("only binary operators can have join selectivity")));
+				 errmsg("only binary operators can have join selectivity")));
 		if (canHash)
 			ereport(ERROR,
 					(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
@@ -450,11 +452,10 @@ OperatorCreate(const char *operatorName,
 	 */
 
 	/*
-	 * Look up registered procedures -- find the return type of
-	 * procedureName to place in "result" field. Do this before shells are
-	 * created so we don't have to worry about deleting them later.
+	 * Look up registered procedures -- find the return type of procedureName
+	 * to place in "result" field. Do this before shells are created so we
+	 * don't have to worry about deleting them later.
 	 */
-	MemSet(typeId, 0, FUNC_MAX_ARGS * sizeof(Oid));
 	if (!OidIsValid(leftTypeId))
 	{
 		typeId[0] = rightTypeId;
@@ -479,7 +480,6 @@ OperatorCreate(const char *operatorName,
 	 */
 	if (restrictionName)
 	{
-		MemSet(typeId, 0, FUNC_MAX_ARGS * sizeof(Oid));
 		typeId[0] = INTERNALOID;	/* Query */
 		typeId[1] = OIDOID;		/* operator OID */
 		typeId[2] = INTERNALOID;	/* args list */
@@ -495,7 +495,6 @@ OperatorCreate(const char *operatorName,
 	 */
 	if (joinName)
 	{
-		MemSet(typeId, 0, FUNC_MAX_ARGS * sizeof(Oid));
 		typeId[0] = INTERNALOID;	/* Query */
 		typeId[1] = OIDOID;		/* operator OID */
 		typeId[2] = INTERNALOID;	/* args list */
@@ -521,7 +520,7 @@ OperatorCreate(const char *operatorName,
 	namestrcpy(&oname, operatorName);
 	values[i++] = NameGetDatum(&oname); /* oprname */
 	values[i++] = ObjectIdGetDatum(operatorNamespace);	/* oprnamespace */
-	values[i++] = Int32GetDatum(GetUserId());	/* oprowner */
+	values[i++] = ObjectIdGetDatum(GetUserId());		/* oprowner */
 	values[i++] = CharGetDatum(leftTypeId ? (rightTypeId ? 'b' : 'r') : 'l');	/* oprkind */
 	values[i++] = BoolGetDatum(canHash);		/* oprcanhash */
 	values[i++] = ObjectIdGetDatum(leftTypeId); /* oprleft */
@@ -622,7 +621,7 @@ OperatorCreate(const char *operatorName,
 	values[i++] = ObjectIdGetDatum(restOid);	/* oprrest */
 	values[i++] = ObjectIdGetDatum(joinOid);	/* oprjoin */
 
-	pg_operator_desc = heap_openr(OperatorRelationName, RowExclusiveLock);
+	pg_operator_desc = heap_open(OperatorRelationId, RowExclusiveLock);
 
 	/*
 	 * If we are adding to an operator shell, update; else insert
@@ -637,7 +636,7 @@ OperatorCreate(const char *operatorName,
 				 operatorObjectId);
 
 		tup = heap_modifytuple(tup,
-							   pg_operator_desc,
+							   RelationGetDescr(pg_operator_desc),
 							   values,
 							   nulls,
 							   replaces);
@@ -656,20 +655,20 @@ OperatorCreate(const char *operatorName,
 	CatalogUpdateIndexes(pg_operator_desc, tup);
 
 	/* Add dependencies for the entry */
-	makeOperatorDependencies(tup, RelationGetRelid(pg_operator_desc));
+	makeOperatorDependencies(tup);
 
 	heap_close(pg_operator_desc, RowExclusiveLock);
 
 	/*
 	 * If a commutator and/or negator link is provided, update the other
-	 * operator(s) to point at this one, if they don't already have a
-	 * link. This supports an alternate style of operator definition
-	 * wherein the user first defines one operator without giving negator
-	 * or commutator, then defines the other operator of the pair with the
-	 * proper commutator or negator attribute.	That style doesn't require
-	 * creation of a shell, and it's the only style that worked right
-	 * before Postgres version 6.5. This code also takes care of the
-	 * situation where the new operator is its own commutator.
+	 * operator(s) to point at this one, if they don't already have a link.
+	 * This supports an alternate style of operator definition wherein the
+	 * user first defines one operator without giving negator or commutator,
+	 * then defines the other operator of the pair with the proper commutator
+	 * or negator attribute.  That style doesn't require creation of a shell,
+	 * and it's the only style that worked right before Postgres version 6.5.
+	 * This code also takes care of the situation where the new operator is
+	 * its own commutator.
 	 */
 	if (selfCommutator)
 		commutatorId = operatorObjectId;
@@ -723,7 +722,7 @@ get_other_operator(List *otherOp, Oid otherLeftTypeId, Oid otherRightTypeId,
 		if (!isCommutator)
 			ereport(ERROR,
 					(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
-					 errmsg("operator cannot be its own negator or sort operator")));
+			 errmsg("operator cannot be its own negator or sort operator")));
 		return InvalidOid;
 	}
 
@@ -775,16 +774,16 @@ OperatorUpd(Oid baseId, Oid commId, Oid negId)
 	 */
 	CommandCounterIncrement();
 
-	pg_operator_desc = heap_openr(OperatorRelationName, RowExclusiveLock);
+	pg_operator_desc = heap_open(OperatorRelationId, RowExclusiveLock);
 
 	tup = SearchSysCacheCopy(OPEROID,
 							 ObjectIdGetDatum(commId),
 							 0, 0, 0);
 
 	/*
-	 * if the commutator and negator are the same operator, do one update.
-	 * XXX this is probably useless code --- I doubt it ever makes sense
-	 * for commutator and negator to be the same thing...
+	 * if the commutator and negator are the same operator, do one update. XXX
+	 * this is probably useless code --- I doubt it ever makes sense for
+	 * commutator and negator to be the same thing...
 	 */
 	if (commId == negId)
 	{
@@ -807,7 +806,7 @@ OperatorUpd(Oid baseId, Oid commId, Oid negId)
 				}
 
 				tup = heap_modifytuple(tup,
-									   pg_operator_desc,
+									   RelationGetDescr(pg_operator_desc),
 									   values,
 									   nulls,
 									   replaces);
@@ -832,7 +831,7 @@ OperatorUpd(Oid baseId, Oid commId, Oid negId)
 		replaces[Anum_pg_operator_oprcom - 1] = 'r';
 
 		tup = heap_modifytuple(tup,
-							   pg_operator_desc,
+							   RelationGetDescr(pg_operator_desc),
 							   values,
 							   nulls,
 							   replaces);
@@ -858,7 +857,7 @@ OperatorUpd(Oid baseId, Oid commId, Oid negId)
 		replaces[Anum_pg_operator_oprnegate - 1] = 'r';
 
 		tup = heap_modifytuple(tup,
-							   pg_operator_desc,
+							   RelationGetDescr(pg_operator_desc),
 							   values,
 							   nulls,
 							   replaces);
@@ -879,23 +878,24 @@ OperatorUpd(Oid baseId, Oid commId, Oid negId)
  * the given operator is a shell.
  */
 static void
-makeOperatorDependencies(HeapTuple tuple, Oid pg_operator_relid)
+makeOperatorDependencies(HeapTuple tuple)
 {
 	Form_pg_operator oper = (Form_pg_operator) GETSTRUCT(tuple);
 	ObjectAddress myself,
 				referenced;
 
-	myself.classId = pg_operator_relid;
+	myself.classId = OperatorRelationId;
 	myself.objectId = HeapTupleGetOid(tuple);
 	myself.objectSubId = 0;
 
 	/* In case we are updating a shell, delete any existing entries */
 	deleteDependencyRecordsFor(myself.classId, myself.objectId);
+	deleteSharedDependencyRecordsFor(myself.classId, myself.objectId);
 
 	/* Dependency on namespace */
 	if (OidIsValid(oper->oprnamespace))
 	{
-		referenced.classId = get_system_catalog_relid(NamespaceRelationName);
+		referenced.classId = NamespaceRelationId;
 		referenced.objectId = oper->oprnamespace;
 		referenced.objectSubId = 0;
 		recordDependencyOn(&myself, &referenced, DEPENDENCY_NORMAL);
@@ -904,7 +904,7 @@ makeOperatorDependencies(HeapTuple tuple, Oid pg_operator_relid)
 	/* Dependency on left type */
 	if (OidIsValid(oper->oprleft))
 	{
-		referenced.classId = RelOid_pg_type;
+		referenced.classId = TypeRelationId;
 		referenced.objectId = oper->oprleft;
 		referenced.objectSubId = 0;
 		recordDependencyOn(&myself, &referenced, DEPENDENCY_NORMAL);
@@ -913,7 +913,7 @@ makeOperatorDependencies(HeapTuple tuple, Oid pg_operator_relid)
 	/* Dependency on right type */
 	if (OidIsValid(oper->oprright))
 	{
-		referenced.classId = RelOid_pg_type;
+		referenced.classId = TypeRelationId;
 		referenced.objectId = oper->oprright;
 		referenced.objectSubId = 0;
 		recordDependencyOn(&myself, &referenced, DEPENDENCY_NORMAL);
@@ -922,7 +922,7 @@ makeOperatorDependencies(HeapTuple tuple, Oid pg_operator_relid)
 	/* Dependency on result type */
 	if (OidIsValid(oper->oprresult))
 	{
-		referenced.classId = RelOid_pg_type;
+		referenced.classId = TypeRelationId;
 		referenced.objectId = oper->oprresult;
 		referenced.objectSubId = 0;
 		recordDependencyOn(&myself, &referenced, DEPENDENCY_NORMAL);
@@ -932,16 +932,16 @@ makeOperatorDependencies(HeapTuple tuple, Oid pg_operator_relid)
 	 * NOTE: we do not consider the operator to depend on the associated
 	 * operators oprcom, oprnegate, oprlsortop, oprrsortop, oprltcmpop,
 	 * oprgtcmpop.	We would not want to delete this operator if those go
-	 * away, but only reset the link fields; which is not a function that
-	 * the dependency code can presently handle.  (Something could perhaps
-	 * be done with objectSubId though.)  For now, it's okay to let those
-	 * links dangle if a referenced operator is removed.
+	 * away, but only reset the link fields; which is not a function that the
+	 * dependency code can presently handle.  (Something could perhaps be done
+	 * with objectSubId though.)  For now, it's okay to let those links dangle
+	 * if a referenced operator is removed.
 	 */
 
 	/* Dependency on implementation function */
 	if (OidIsValid(oper->oprcode))
 	{
-		referenced.classId = RelOid_pg_proc;
+		referenced.classId = ProcedureRelationId;
 		referenced.objectId = oper->oprcode;
 		referenced.objectSubId = 0;
 		recordDependencyOn(&myself, &referenced, DEPENDENCY_NORMAL);
@@ -950,7 +950,7 @@ makeOperatorDependencies(HeapTuple tuple, Oid pg_operator_relid)
 	/* Dependency on restriction selectivity function */
 	if (OidIsValid(oper->oprrest))
 	{
-		referenced.classId = RelOid_pg_proc;
+		referenced.classId = ProcedureRelationId;
 		referenced.objectId = oper->oprrest;
 		referenced.objectSubId = 0;
 		recordDependencyOn(&myself, &referenced, DEPENDENCY_NORMAL);
@@ -959,9 +959,13 @@ makeOperatorDependencies(HeapTuple tuple, Oid pg_operator_relid)
 	/* Dependency on join selectivity function */
 	if (OidIsValid(oper->oprjoin))
 	{
-		referenced.classId = RelOid_pg_proc;
+		referenced.classId = ProcedureRelationId;
 		referenced.objectId = oper->oprjoin;
 		referenced.objectSubId = 0;
 		recordDependencyOn(&myself, &referenced, DEPENDENCY_NORMAL);
 	}
+
+	/* Dependency on owner */
+	recordDependencyOnOwner(OperatorRelationId, HeapTupleGetOid(tuple),
+							oper->oprowner);
 }
