@@ -5,12 +5,12 @@
  * PostgreSQL Integrated Autovacuum Daemon
  *
  *
- * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2006, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/postmaster/autovacuum.c,v 1.5.2.7 2007/01/14 20:18:30 alvherre Exp $
+ *	  $PostgreSQL: pgsql/src/backend/postmaster/autovacuum.c,v 1.29 2006/11/21 20:59:52 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -23,7 +23,8 @@
 
 #include "access/genam.h"
 #include "access/heapam.h"
-#include "access/xlog.h"
+#include "access/transam.h"
+#include "access/xact.h"
 #include "catalog/indexing.h"
 #include "catalog/namespace.h"
 #include "catalog/pg_autovacuum.h"
@@ -43,11 +44,10 @@
 #include "tcop/tcopprot.h"
 #include "utils/flatfiles.h"
 #include "utils/fmgroids.h"
+#include "utils/lsyscache.h"
 #include "utils/memutils.h"
 #include "utils/ps_status.h"
-#include "utils/lsyscache.h"
-#include "utils/rel.h"
-#include "utils/relcache.h"
+#include "utils/syscache.h"
 
 
 /*
@@ -59,6 +59,7 @@ int			autovacuum_vac_thresh;
 double		autovacuum_vac_scale;
 int			autovacuum_anl_thresh;
 double		autovacuum_anl_scale;
+int			autovacuum_freeze_max_age;
 
 int			autovacuum_vac_cost_delay;
 int			autovacuum_vac_cost_limit;
@@ -70,6 +71,12 @@ static bool am_autovacuum = false;
 static time_t last_autovac_start_time = 0;
 static time_t last_autovac_stop_time = 0;
 
+/* Comparison point for determining whether freeze_max_age is exceeded */
+static TransactionId recentXid;
+
+/* Default freeze_min_age to use for autovacuum (varies by database) */
+static int	default_freeze_min_age;
+
 /* Memory context for long-lived data */
 static MemoryContext AutovacMemCxt;
 
@@ -79,9 +86,7 @@ typedef struct autovac_dbase
 	Oid			oid;
 	char	   *name;
 	TransactionId frozenxid;
-	TransactionId vacuumxid;
 	PgStat_StatDBEntry *entry;
-	int32		age;
 } autovac_dbase;
 
 /* struct to keep track of tables to vacuum and/or analyze */
@@ -91,6 +96,7 @@ typedef struct autovac_table
 	Oid			toastrelid;
 	bool		dovacuum;
 	bool		doanalyze;
+	int			freeze_min_age;
 	int			vacuum_cost_delay;
 	int			vacuum_cost_limit;
 } autovac_table;
@@ -100,7 +106,6 @@ typedef struct autovac_table
 static pid_t autovac_forkexec(void);
 #endif
 NON_EXEC_STATIC void AutoVacMain(int argc, char *argv[]);
-static void process_whole_db(void);
 static void do_autovacuum(PgStat_StatDBEntry *dbentry);
 static List *autovac_get_database_list(void);
 static void test_rel_for_autovac(Oid relid, PgStat_StatTabEntry *tabentry,
@@ -108,10 +113,9 @@ static void test_rel_for_autovac(Oid relid, PgStat_StatTabEntry *tabentry,
 					 Form_pg_autovacuum avForm,
 					 List **vacuum_tables,
 					 List **toast_table_ids);
-static void autovacuum_do_vac_analyze(List *relids, bool dovacuum,
-						  bool doanalyze, bool freeze);
-static void autovac_report_activity(VacuumStmt *vacstmt,
-						List *relids);
+static void autovacuum_do_vac_analyze(Oid relid, bool dovacuum,
+						  bool doanalyze, int freeze_min_age);
+static void autovac_report_activity(VacuumStmt *vacstmt, Oid relid);
 
 
 /*
@@ -125,10 +129,6 @@ autovac_start(void)
 	time_t		curtime;
 	pid_t		AutoVacPID;
 
-	/* Do nothing if no autovacuum process needed */
-	if (!AutoVacuumingActive())
-		return 0;
-
 	/*
 	 * Do nothing if too soon since last autovacuum exit.  This limits how
 	 * often the daemon runs.  Since the time per iteration can be quite
@@ -143,6 +143,7 @@ autovac_start(void)
 	 *
 	 * XXX todo: implement sleep scale factor that existed in contrib code.
 	 */
+
 	curtime = time(NULL);
 	if ((unsigned int) (curtime - last_autovac_stop_time) <
 		(unsigned int) autovacuum_naptime)
@@ -171,6 +172,9 @@ autovac_start(void)
 			/* Close the postmaster's sockets */
 			ClosePostmasterPorts(false);
 
+			/* Lose the postmaster's on-exit routines */
+			on_exit_reset();
+
 			AutoVacMain(0, NULL);
 			break;
 #endif
@@ -204,7 +208,7 @@ autovac_forkexec(void)
 	int			ac = 0;
 
 	av[ac++] = "postgres";
-	av[ac++] = "-forkautovac";
+	av[ac++] = "--forkautovac";
 	av[ac++] = NULL;			/* filled in by postmaster_forkexec */
 	av[ac] = NULL;
 
@@ -222,9 +226,9 @@ AutoVacMain(int argc, char *argv[])
 {
 	ListCell   *cell;
 	List	   *dblist;
-	TransactionId nextXid;
 	autovac_dbase *db;
-	bool		whole_db;
+	TransactionId xidForceLimit;
+	bool		for_xid_wrap;
 	sigjmp_buf	local_sigjmp_buf;
 
 	/* we are a postmaster subprocess now */
@@ -234,15 +238,22 @@ AutoVacMain(int argc, char *argv[])
 	/* reset MyProcPid */
 	MyProcPid = getpid();
 
-	/* Lose the postmaster's on-exit routines */
-	on_exit_reset();
-
 	/* Identify myself via ps */
-	init_ps_display("autovacuum process", "", "");
-	set_ps_display("");
+	init_ps_display("autovacuum process", "", "", "");
 
 	SetProcessingMode(InitProcessing);
 
+	/*
+	 * If possible, make this process a group leader, so that the postmaster
+	 * can signal any child processes too.  (autovacuum probably never has
+	 * any child processes, but for consistency we make all postmaster
+	 * child processes do this.)
+	 */
+#ifdef HAVE_SETSID
+	if (setsid() < 0)
+		elog(FATAL, "setsid() failed: %m");
+#endif
+
 	/*
 	 * Set up signal handlers.	We operate on databases much like a regular
 	 * backend, so we use the same signal handling.  See equivalent code in
@@ -272,6 +283,16 @@ AutoVacMain(int argc, char *argv[])
 	/* Early initialization */
 	BaseInit();
 
+	/*
+	 * Create a per-backend PGPROC struct in shared memory, except in the
+	 * EXEC_BACKEND case where this was done in SubPostmasterMain. We must do
+	 * this before we can use LWLocks (and in the EXEC_BACKEND case we already
+	 * had to do some stuff with LWLocks).
+	 */
+#ifndef EXEC_BACKEND
+	InitProcess();
+#endif
+
 	/*
 	 * If an exception is encountered, processing resumes here.
 	 *
@@ -298,19 +319,33 @@ AutoVacMain(int argc, char *argv[])
 
 	PG_SETMASK(&UnBlockSig);
 
+	/*
+	 * Force zero_damaged_pages OFF in the autovac process, even if it is set
+	 * in postgresql.conf.	We don't really want such a dangerous option being
+	 * applied non-interactively.
+	 */
+	SetConfigOption("zero_damaged_pages", "false", PGC_SUSET, PGC_S_OVERRIDE);
+
 	/* Get a list of databases */
 	dblist = autovac_get_database_list();
 
 	/*
-	 * Get the next Xid that was current as of the last checkpoint. We need it
-	 * to determine whether databases are about to need database-wide vacuums.
+	 * Determine the oldest datfrozenxid/relfrozenxid that we will allow
+	 * to pass without forcing a vacuum.  (This limit can be tightened for
+	 * particular tables, but not loosened.)
 	 */
-	nextXid = GetRecentNextXid();
+	recentXid = ReadNewTransactionId();
+	xidForceLimit = recentXid - autovacuum_freeze_max_age;
+	/* ensure it's a "normal" XID, else TransactionIdPrecedes misbehaves */
+	if (xidForceLimit < FirstNormalTransactionId)
+		xidForceLimit -= FirstNormalTransactionId;
 
 	/*
 	 * Choose a database to connect to.  We pick the database that was least
-	 * recently auto-vacuumed, or one that needs database-wide vacuum (to
-	 * prevent Xid wraparound-related data loss).
+	 * recently auto-vacuumed, or one that needs vacuuming to prevent Xid
+	 * wraparound-related data loss.  If any db at risk of wraparound is
+	 * found, we pick the one with oldest datfrozenxid,
+	 * independently of autovacuum times.
 	 *
 	 * Note that a database with no stats entry is not considered, except for
 	 * Xid wraparound purposes.  The theory is that if no one has ever
@@ -325,65 +360,36 @@ AutoVacMain(int argc, char *argv[])
 	 * starvation for less busy databases.
 	 */
 	db = NULL;
-	whole_db = false;
-
+	for_xid_wrap = false;
 	foreach(cell, dblist)
 	{
 		autovac_dbase *tmp = lfirst(cell);
-		bool		this_whole_db;
-		int32		freeze_age,
-					vacuum_age;
 
-		/*
-		 * We look for the database that most urgently needs a database-wide
-		 * vacuum.	We decide that a database-wide vacuum is needed 100000
-		 * transactions sooner than vacuum.c's vac_truncate_clog() would
-		 * decide to start giving warnings.  If any such db is found, we
-		 * ignore all other dbs.
-		 *
-		 * Unlike vacuum.c, we also look at vacuumxid.	This is so that
-		 * pg_clog can be kept trimmed to a reasonable size.
-		 */
-		freeze_age = (int32) (nextXid - tmp->frozenxid);
-		vacuum_age = (int32) (nextXid - tmp->vacuumxid);
-		tmp->age = Max(freeze_age, vacuum_age);
+		/* Find pgstat entry if any */
+		tmp->entry = pgstat_fetch_stat_dbentry(tmp->oid);
 
-		this_whole_db = (tmp->age >
-						 (int32) ((MaxTransactionId >> 3) * 3 - 100000));
-		if (whole_db || this_whole_db)
+		/* Check to see if this one is at risk of wraparound */
+		if (TransactionIdPrecedes(tmp->frozenxid, xidForceLimit))
 		{
-			if (!this_whole_db)
-				continue;
-			if (db == NULL || tmp->age > db->age)
-			{
+			if (db == NULL ||
+				TransactionIdPrecedes(tmp->frozenxid, db->frozenxid))
 				db = tmp;
-				whole_db = true;
-			}
+			for_xid_wrap = true;
 			continue;
 		}
+		else if (for_xid_wrap)
+			continue;			/* ignore not-at-risk DBs */
 
 		/*
-		 * Otherwise, skip a database with no pgstat entry; it means it hasn't
-		 * seen any activity.
+		 * Otherwise, skip a database with no pgstat entry; it means it
+		 * hasn't seen any activity.
 		 */
-		tmp->entry = pgstat_fetch_stat_dbentry(tmp->oid);
 		if (!tmp->entry)
 			continue;
 
 		/*
-		 * Don't try to access a database that was dropped.  This could only
-		 * happen if we read the pg_database flat file right before it was
-		 * modified, after the database was dropped from the pg_database
-		 * table.  (This is of course a not-very-bulletproof test, but it's
-		 * cheap to make.  If we do mistakenly choose a recently dropped
-		 * database, InitPostgres will fail and we'll drop out until the next
-		 * autovac run.)
-		 */
-		if (tmp->entry->destroy != 0)
-			continue;
-
-		/*
-		 * Else remember the db with oldest autovac time.
+		 * Remember the db with oldest autovac time.  (If we are here,
+		 * both tmp->entry and db->entry must be non-null.)
 		 */
 		if (db == NULL ||
 			tmp->entry->last_autovac_time < db->entry->last_autovac_time)
@@ -404,11 +410,14 @@ AutoVacMain(int argc, char *argv[])
 
 		/*
 		 * Connect to the selected database
+		 *
+		 * Note: if we have selected a just-deleted database (due to using
+		 * stale stats info), we'll fail and exit here.
 		 */
 		InitPostgres(db->name, NULL);
 		SetProcessingMode(NormalProcessing);
-		set_ps_display(db->name);
-		ereport(LOG,
+		set_ps_display(db->name, false);
+		ereport(DEBUG1,
 				(errmsg("autovacuum: processing database \"%s\"", db->name)));
 
 		/* Create the memory context where cross-transaction state is stored */
@@ -421,10 +430,7 @@ AutoVacMain(int argc, char *argv[])
 		/*
 		 * And do an appropriate amount of work
 		 */
-		if (whole_db)
-			process_whole_db();
-		else
-			do_autovacuum(db->entry);
+		do_autovacuum(db->entry);
 	}
 
 	/* One iteration done, go away */
@@ -447,7 +453,6 @@ autovac_get_database_list(void)
 	Oid			db_id;
 	Oid			db_tablespace;
 	TransactionId db_frozenxid;
-	TransactionId db_vacuumxid;
 
 	filename = database_getflatfilename();
 	db_file = AllocateFile(filename, "r");
@@ -457,8 +462,7 @@ autovac_get_database_list(void)
 				 errmsg("could not open file \"%s\": %m", filename)));
 
 	while (read_pg_database_line(db_file, thisname, &db_id,
-								 &db_tablespace, &db_frozenxid,
-								 &db_vacuumxid))
+								 &db_tablespace, &db_frozenxid))
 	{
 		autovac_dbase *db;
 
@@ -467,10 +471,8 @@ autovac_get_database_list(void)
 		db->oid = db_id;
 		db->name = pstrdup(thisname);
 		db->frozenxid = db_frozenxid;
-		db->vacuumxid = db_vacuumxid;
-		/* these get set later: */
+		/* this gets set later: */
 		db->entry = NULL;
-		db->age = 0;
 
 		dblist = lappend(dblist, db);
 	}
@@ -481,90 +483,12 @@ autovac_get_database_list(void)
 	return dblist;
 }
 
-/*
- * Return a palloc'ed copy of the pg_database entry for the given database.
- * Note that no lock is retained on the entry whatsoever, so it may be stale by
- * the time the caller inspects it.  This is sufficient for our purposes
- * however.
- */
-static Form_pg_database
-get_pg_database_entry(Oid dbid)
-{
-	Form_pg_database dbForm;
-	ScanKeyData entry[1];
-	Relation	dbRel;
-	SysScanDesc scan;
-	HeapTuple	tup;
-
-	dbRel = heap_open(DatabaseRelationId, AccessShareLock);
-
-	/* Must use a table scan, since there's no syscache for pg_database */
-	ScanKeyInit(&entry[0],
-				ObjectIdAttributeNumber,
-				BTEqualStrategyNumber, F_OIDEQ,
-				ObjectIdGetDatum(MyDatabaseId));
-
-	scan = systable_beginscan(dbRel, DatabaseOidIndexId, true,
-							  SnapshotNow, 1, entry);
-
-	tup = systable_getnext(scan);
-
-	if (!HeapTupleIsValid(tup))
-		elog(ERROR, "could not find tuple for database %u", MyDatabaseId);
-
-	dbForm = (Form_pg_database) palloc(sizeof(FormData_pg_database));
-	memcpy(dbForm, GETSTRUCT(tup), sizeof(FormData_pg_database));
-
-	systable_endscan(scan);
-
-	heap_close(dbRel, AccessShareLock);
-
-	return dbForm;
-}
-
-/*
- * Process a whole database.  If it's a template database or is disallowing
- * connection by means of datallowconn=false, then issue a VACUUM FREEZE.
- * Else use a plain VACUUM.
- */
-static void
-process_whole_db(void)
-{
-	Form_pg_database dbForm;
-	bool		freeze;
-
-	/* Start a transaction so our commands have one to play into. */
-	StartTransactionCommand();
-
-	 /* functions in indexes may want a snapshot set */
-	ActiveSnapshot = CopySnapshot(GetTransactionSnapshot());
-
-	/*
-	 * Clean up any dead statistics collector entries for this DB.
-	 */
-	pgstat_vacuum_tabstat();
-
-	dbForm = get_pg_database_entry(MyDatabaseId);
-
-	freeze = (!dbForm->datallowconn || dbForm->datistemplate);
-
-	pfree(dbForm);
-
-	elog(DEBUG2, "autovacuum: VACUUM%s whole database",
-		 (freeze) ? " FREEZE" : "");
-
-	autovacuum_do_vac_analyze(NIL, true, false, freeze);
-
-	/* Finally close out the last transaction. */
-	CommitTransactionCommand();
-}
-
 /*
  * Process a database table-by-table
  *
- * dbentry must be a valid pointer to the database entry in the stats
- * databases' hash table, and it will be used to determine whether vacuum or
- * analyze is needed on a per-table basis.
+ * dbentry is either a pointer to the database entry in the stats databases
+ * hash table, or NULL if we couldn't find any entry (the latter case occurs
+ * only if we are forcing a vacuum for anti-wrap purposes).
  *
  * Note that CHECK_FOR_INTERRUPTS is supposed to be used in certain spots in
  * order not to ignore shutdown commands for too long.
@@ -576,33 +500,43 @@ do_autovacuum(PgStat_StatDBEntry *dbentry)
 				avRel;
 	HeapTuple	tuple;
 	HeapScanDesc relScan;
+	Form_pg_database dbForm;
 	List	   *vacuum_tables = NIL;
 	List	   *toast_table_ids = NIL;
 	ListCell   *cell;
 	PgStat_StatDBEntry *shared;
-	Form_pg_database dbForm;
-	bool		istemplate;
 
 	/* Start a transaction so our commands have one to play into. */
 	StartTransactionCommand();
 
-	 /* functions in indexes may want a snapshot set */
+	/* functions in indexes may want a snapshot set */
 	ActiveSnapshot = CopySnapshot(GetTransactionSnapshot());
 
 	/*
-	 * Clean up any dead statistics collector entries for this DB.
-	 * We always want to do this exactly once per DB-processing cycle,
-	 * even if we find nothing worth vacuuming in the database.
+	 * Clean up any dead statistics collector entries for this DB. We always
+	 * want to do this exactly once per DB-processing cycle, even if we find
+	 * nothing worth vacuuming in the database.
 	 */
 	pgstat_vacuum_tabstat();
 
 	/*
-	 * In a template database, we need to avoid putting our Xid in any table,
-	 * so disallow analyzes and force use of VACUUM FREEZE.
+	 * Find the pg_database entry and select the default freeze_min_age.
+	 * We use zero in template and nonconnectable databases,
+	 * else the system-wide default.
 	 */
-	dbForm = get_pg_database_entry(MyDatabaseId);
-	istemplate = (!dbForm->datallowconn || dbForm->datistemplate);
-	pfree(dbForm);
+	tuple = SearchSysCache(DATABASEOID,
+						   ObjectIdGetDatum(MyDatabaseId),
+						   0, 0, 0);
+	if (!HeapTupleIsValid(tuple))
+		elog(ERROR, "cache lookup failed for database %u", MyDatabaseId);
+	dbForm = (Form_pg_database) GETSTRUCT(tuple);
+
+	if (dbForm->datistemplate || !dbForm->datallowconn)
+		default_freeze_min_age = 0;
+	else
+		default_freeze_min_age = vacuum_freeze_min_age;
+
+	ReleaseSysCache(tuple);
 
 	/*
 	 * StartTransactionCommand and CommitTransactionCommand will automatically
@@ -677,9 +611,11 @@ do_autovacuum(PgStat_StatDBEntry *dbentry)
 		if (classForm->relisshared && PointerIsValid(shared))
 			tabentry = hash_search(shared->tables, &relid,
 								   HASH_FIND, NULL);
-		else
+		else if (PointerIsValid(dbentry))
 			tabentry = hash_search(dbentry->tables, &relid,
 								   HASH_FIND, NULL);
+		else
+			tabentry = NULL;
 
 		test_rel_for_autovac(relid, tabentry, classForm, avForm,
 							 &vacuum_tables, &toast_table_ids);
@@ -720,13 +656,18 @@ do_autovacuum(PgStat_StatDBEntry *dbentry)
 		VacuumCostDelay = tab->vacuum_cost_delay;
 		VacuumCostLimit = tab->vacuum_cost_limit;
 
-		/* in a template database, we never analyze and force freezing */
-		autovacuum_do_vac_analyze(list_make1_oid(tab->relid),
+		autovacuum_do_vac_analyze(tab->relid,
 								  tab->dovacuum,
-								  !istemplate && tab->doanalyze,
-								  istemplate);
+								  tab->doanalyze,
+								  tab->freeze_min_age);
 	}
 
+	/*
+	 * Update pg_database.datfrozenxid, and truncate pg_clog if possible.
+	 * We only need to do this once, not after each table.
+	 */
+	vac_update_datfrozenxid();
+
 	/* Finally close out the last transaction. */
 	CommitTransactionCommand();
 }
@@ -748,10 +689,13 @@ do_autovacuum(PgStat_StatDBEntry *dbentry)
  * the number of tuples (both live and dead) that there were as of the last
  * analyze.  This is asymmetric to the VACUUM case.
  *
+ * We also force vacuum if the table's relfrozenxid is more than freeze_max_age
+ * transactions back.
+ *
  * A table whose pg_autovacuum.enabled value is false, is automatically
- * skipped.  Thus autovacuum can be disabled for specific tables.  Also,
- * when the stats collector does not have data about a table, it will be
- * skipped.
+ * skipped (unless we have to vacuum it due to freeze_max_age).  Thus
+ * autovacuum can be disabled for specific tables.  Also, when the stats
+ * collector does not have data about a table, it will be skipped.
  *
  * A table whose vac_base_thresh value is <0 takes the base value from the
  * autovacuum_vacuum_threshold GUC variable.  Similarly, a vac_scale_factor
@@ -765,50 +709,28 @@ test_rel_for_autovac(Oid relid, PgStat_StatTabEntry *tabentry,
 					 List **vacuum_tables,
 					 List **toast_table_ids)
 {
-	Relation	rel;
+	bool		force_vacuum;
+	bool		dovacuum;
+	bool		doanalyze;
 	float4		reltuples;		/* pg_class.reltuples */
-
 	/* constants from pg_autovacuum or GUC variables */
 	int			vac_base_thresh,
 				anl_base_thresh;
 	float4		vac_scale_factor,
 				anl_scale_factor;
-
 	/* thresholds calculated from above constants */
 	float4		vacthresh,
 				anlthresh;
-
 	/* number of vacuum (resp. analyze) tuples at this time */
 	float4		vactuples,
 				anltuples;
-
+	/* freeze parameters */
+	int			freeze_min_age;
+	int			freeze_max_age;
+	TransactionId xidForceLimit;
 	/* cost-based vacuum delay parameters */
 	int			vac_cost_limit;
 	int			vac_cost_delay;
-	bool		dovacuum;
-	bool		doanalyze;
-
-	/* User disabled it in pg_autovacuum? */
-	if (avForm && !avForm->enabled)
-		return;
-
-	/*
-	 * Skip a table not found in stat hash.  If it's not acted upon, there's
-	 * no need to vacuum it.  (Note that database-level check will take care
-	 * of Xid wraparound.)
-	 */
-	if (!PointerIsValid(tabentry))
-		return;
-
-	rel = RelationIdGetRelation(relid);
-	/* The table was recently dropped? */
-	if (!PointerIsValid(rel))
-		return;
-
-	reltuples = rel->rd_rel->reltuples;
-	vactuples = tabentry->n_dead_tuples;
-	anltuples = tabentry->n_live_tuples + tabentry->n_dead_tuples -
-		tabentry->last_anl_tuples;
 
 	/*
 	 * If there is a tuple in pg_autovacuum, use it; else, use the GUC
@@ -827,6 +749,12 @@ test_rel_for_autovac(Oid relid, PgStat_StatTabEntry *tabentry,
 		anl_base_thresh = (avForm->anl_base_thresh >= 0) ?
 			avForm->anl_base_thresh : autovacuum_anl_thresh;
 
+		freeze_min_age = (avForm->freeze_min_age >= 0) ?
+			avForm->freeze_min_age : default_freeze_min_age;
+		freeze_max_age = (avForm->freeze_max_age >= 0) ?
+			Min(avForm->freeze_max_age, autovacuum_freeze_max_age) :
+			autovacuum_freeze_max_age;
+
 		vac_cost_limit = (avForm->vac_cost_limit >= 0) ?
 			avForm->vac_cost_limit :
 			((autovacuum_vac_cost_limit >= 0) ?
@@ -845,6 +773,9 @@ test_rel_for_autovac(Oid relid, PgStat_StatTabEntry *tabentry,
 		anl_scale_factor = autovacuum_anl_scale;
 		anl_base_thresh = autovacuum_anl_thresh;
 
+		freeze_min_age = default_freeze_min_age;
+		freeze_max_age = autovacuum_freeze_max_age;
+
 		vac_cost_limit = (autovacuum_vac_cost_limit >= 0) ?
 			autovacuum_vac_cost_limit : VacuumCostLimit;
 
@@ -852,22 +783,51 @@ test_rel_for_autovac(Oid relid, PgStat_StatTabEntry *tabentry,
 			autovacuum_vac_cost_delay : VacuumCostDelay;
 	}
 
-	vacthresh = (float4) vac_base_thresh + vac_scale_factor * reltuples;
-	anlthresh = (float4) anl_base_thresh + anl_scale_factor * reltuples;
+	/* Force vacuum if table is at risk of wraparound */
+	xidForceLimit = recentXid - freeze_max_age;
+	if (xidForceLimit < FirstNormalTransactionId)
+		xidForceLimit -= FirstNormalTransactionId;
+	force_vacuum = (TransactionIdIsNormal(classForm->relfrozenxid) &&
+					TransactionIdPrecedes(classForm->relfrozenxid,
+										  xidForceLimit));
 
-	/*
-	 * Note that we don't need to take special consideration for stat reset,
-	 * because if that happens, the last vacuum and analyze counts will be
-	 * reset too.
-	 */
+	/* User disabled it in pg_autovacuum?  (But ignore if at risk) */
+	if (avForm && !avForm->enabled && !force_vacuum)
+		return;
 
-	elog(DEBUG3, "%s: vac: %.0f (threshold %.0f), anl: %.0f (threshold %.0f)",
-		 RelationGetRelationName(rel),
-		 vactuples, vacthresh, anltuples, anlthresh);
+	if (PointerIsValid(tabentry))
+	{
+		reltuples = classForm->reltuples;
+		vactuples = tabentry->n_dead_tuples;
+		anltuples = tabentry->n_live_tuples + tabentry->n_dead_tuples -
+			tabentry->last_anl_tuples;
 
-	/* Determine if this table needs vacuum or analyze. */
-	dovacuum = (vactuples > vacthresh);
-	doanalyze = (anltuples > anlthresh);
+		vacthresh = (float4) vac_base_thresh + vac_scale_factor * reltuples;
+		anlthresh = (float4) anl_base_thresh + anl_scale_factor * reltuples;
+
+		/*
+		 * Note that we don't need to take special consideration for stat
+		 * reset, because if that happens, the last vacuum and analyze counts
+		 * will be reset too.
+		 */
+		elog(DEBUG3, "%s: vac: %.0f (threshold %.0f), anl: %.0f (threshold %.0f)",
+			 NameStr(classForm->relname),
+			 vactuples, vacthresh, anltuples, anlthresh);
+
+		/* Determine if this table needs vacuum or analyze. */
+		dovacuum = force_vacuum || (vactuples > vacthresh);
+		doanalyze = (anltuples > anlthresh);
+	}
+	else
+	{
+		/*
+		 * Skip a table not found in stat hash, unless we have to force
+		 * vacuum for anti-wrap purposes.  If it's not acted upon, there's
+		 * no need to vacuum it.
+		 */
+		dovacuum = force_vacuum;
+		doanalyze = false;
+	}
 
 	/* ANALYZE refuses to work with pg_statistics */
 	if (relid == StatisticRelationId)
@@ -881,7 +841,7 @@ test_rel_for_autovac(Oid relid, PgStat_StatTabEntry *tabentry,
 			elog(DEBUG2, "autovac: will%s%s %s",
 				 (dovacuum ? " VACUUM" : ""),
 				 (doanalyze ? " ANALYZE" : ""),
-				 RelationGetRelationName(rel));
+				 NameStr(classForm->relname));
 
 		/*
 		 * we must record tables that have a toast table, even if we currently
@@ -896,6 +856,7 @@ test_rel_for_autovac(Oid relid, PgStat_StatTabEntry *tabentry,
 			tab->toastrelid = classForm->reltoastrelid;
 			tab->dovacuum = dovacuum;
 			tab->doanalyze = doanalyze;
+			tab->freeze_min_age = freeze_min_age;
 			tab->vacuum_cost_limit = vac_cost_limit;
 			tab->vacuum_cost_delay = vac_cost_delay;
 
@@ -908,17 +869,15 @@ test_rel_for_autovac(Oid relid, PgStat_StatTabEntry *tabentry,
 		if (dovacuum)
 			*toast_table_ids = lappend_oid(*toast_table_ids, relid);
 	}
-
-	RelationClose(rel);
 }
 
 /*
  * autovacuum_do_vac_analyze
- *		Vacuum and/or analyze a list of tables; or all tables if relids = NIL
+ *		Vacuum and/or analyze the specified table
  */
 static void
-autovacuum_do_vac_analyze(List *relids, bool dovacuum, bool doanalyze,
-						  bool freeze)
+autovacuum_do_vac_analyze(Oid relid, bool dovacuum, bool doanalyze,
+						  int freeze_min_age)
 {
 	VacuumStmt *vacstmt;
 	MemoryContext old_cxt;
@@ -942,15 +901,15 @@ autovacuum_do_vac_analyze(List *relids, bool dovacuum, bool doanalyze,
 	vacstmt->vacuum = dovacuum;
 	vacstmt->full = false;
 	vacstmt->analyze = doanalyze;
-	vacstmt->freeze = freeze;
+	vacstmt->freeze_min_age = freeze_min_age;
 	vacstmt->verbose = false;
-	vacstmt->relation = NULL;	/* all tables, or not used if relids != NIL */
+	vacstmt->relation = NULL;	/* not used since we pass relids list */
 	vacstmt->va_cols = NIL;
 
 	/* Let pgstat know what we're doing */
-	autovac_report_activity(vacstmt, relids);
+	autovac_report_activity(vacstmt, relid);
 
-	vacuum(vacstmt, relids);
+	vacuum(vacstmt, list_make1_oid(relid));
 
 	pfree(vacstmt);
 	MemoryContextSwitchTo(old_cxt);
@@ -958,7 +917,7 @@ autovacuum_do_vac_analyze(List *relids, bool dovacuum, bool doanalyze,
 
 /*
  * autovac_report_activity
- * 		Report to pgstat what autovacuum is doing
+ *		Report to pgstat what autovacuum is doing
  *
  * We send a SQL string corresponding to what the user would see if the
  * equivalent command was to be issued manually.
@@ -967,50 +926,36 @@ autovacuum_do_vac_analyze(List *relids, bool dovacuum, bool doanalyze,
  * done with the current one, and exiting right after the last one, so we don't
  * bother to report "<IDLE>" or some such.
  */
-#define MAX_AUTOVAC_ACTIV_LEN (NAMEDATALEN * 2 + 32)
 static void
-autovac_report_activity(VacuumStmt *vacstmt, List *relids)
+autovac_report_activity(VacuumStmt *vacstmt, Oid relid)
 {
+	char	   *relname = get_rel_name(relid);
+	char	   *nspname = get_namespace_name(get_rel_namespace(relid));
+#define MAX_AUTOVAC_ACTIV_LEN (NAMEDATALEN * 2 + 32)
 	char		activity[MAX_AUTOVAC_ACTIV_LEN];
 
-	/*
-	 * This case is not currently exercised by the autovac code.  Fill it in
-	 * if needed.
-	 */
-	if (list_length(relids) > 1)
-		elog(WARNING, "vacuuming >1 rel unsupported");
-
 	/* Report the command and possible options */
 	if (vacstmt->vacuum)
 		snprintf(activity, MAX_AUTOVAC_ACTIV_LEN,
-					   "VACUUM%s%s%s",
-					   vacstmt->full ? " FULL" : "",
-					   vacstmt->freeze ? " FREEZE" : "",
-					   vacstmt->analyze ? " ANALYZE" : "");
-	else if (vacstmt->analyze)
+				 "VACUUM%s",
+				 vacstmt->analyze ? " ANALYZE" : "");
+	else
 		snprintf(activity, MAX_AUTOVAC_ACTIV_LEN,
-					   "ANALYZE");
+				 "ANALYZE");
 
-	/* Report the qualified name of the first relation, if any */
-	if (list_length(relids) > 0)
+	/*
+	 * Report the qualified name of the relation.
+	 *
+	 * Paranoia is appropriate here in case relation was recently dropped
+	 * --- the lsyscache routines we just invoked will return NULL rather
+	 * than failing.
+	 */
+	if (relname && nspname)
 	{
-		Oid			relid = linitial_oid(relids);
-		Relation	rel;
+		int			len = strlen(activity);
 
-		rel = RelationIdGetRelation(relid);
-		if (rel == NULL)
-			elog(WARNING, "cache lookup failed for relation %u", relid);
-		else
-		{
-			char   *nspname = get_namespace_name(RelationGetNamespace(rel));
-			int		len = strlen(activity);
-
-			snprintf(activity + len, MAX_AUTOVAC_ACTIV_LEN - len,
-					 " %s.%s", nspname, RelationGetRelationName(rel));
-
-			pfree(nspname);
-			RelationClose(rel);
-		}
+		snprintf(activity + len, MAX_AUTOVAC_ACTIV_LEN - len,
+				 " %s.%s", nspname, relname);
 	}
 
 	pgstat_report_activity(activity);
