@@ -3,26 +3,24 @@
  * rewriteDefine.c
  *	  routines for defining a rewrite rule
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2006, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/rewrite/rewriteDefine.c,v 1.90 2003/09/29 00:05:25 petere Exp $
+ *	  $PostgreSQL: pgsql/src/backend/rewrite/rewriteDefine.c,v 1.115 2006/10/04 00:29:56 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
 #include "postgres.h"
 
 #include "access/heapam.h"
-#include "catalog/catname.h"
 #include "catalog/dependency.h"
 #include "catalog/indexing.h"
 #include "catalog/pg_rewrite.h"
-#include "commands/view.h"
 #include "miscadmin.h"
 #include "optimizer/clauses.h"
-#include "parser/parse_relation.h"
+#include "parser/parse_expr.h"
 #include "rewrite/rewriteDefine.h"
 #include "rewrite/rewriteManip.h"
 #include "rewrite/rewriteSupport.h"
@@ -33,8 +31,10 @@
 #include "utils/syscache.h"
 
 
-static void setRuleCheckAsUser(Query *qry, AclId userid);
+static void checkRuleResultList(List *targetList, TupleDesc resultDesc,
+					bool isSelect);
 static bool setRuleCheckAsUser_walker(Node *node, Oid *context);
+static void setRuleCheckAsUser_Query(Query *qry, Oid userid);
 
 
 /*
@@ -85,7 +85,7 @@ InsertRule(char *rulname,
 	/*
 	 * Ready to store new pg_rewrite tuple
 	 */
-	pg_rewrite_desc = heap_openr(RewriteRelationName, RowExclusiveLock);
+	pg_rewrite_desc = heap_open(RewriteRelationId, RowExclusiveLock);
 
 	/*
 	 * Check to see if we are replacing an existing tuple
@@ -100,8 +100,8 @@ InsertRule(char *rulname,
 		if (!replace)
 			ereport(ERROR,
 					(errcode(ERRCODE_DUPLICATE_OBJECT),
-				 errmsg("rule \"%s\" for relation \"%s\" already exists",
-						rulname, get_rel_name(eventrel_oid))));
+					 errmsg("rule \"%s\" for relation \"%s\" already exists",
+							rulname, get_rel_name(eventrel_oid))));
 
 		/*
 		 * When replacing, we don't need to replace every attribute
@@ -113,7 +113,7 @@ InsertRule(char *rulname,
 		replaces[Anum_pg_rewrite_ev_qual - 1] = 'r';
 		replaces[Anum_pg_rewrite_ev_action - 1] = 'r';
 
-		tup = heap_modifytuple(oldtup, pg_rewrite_desc,
+		tup = heap_modifytuple(oldtup, RelationGetDescr(pg_rewrite_desc),
 							   values, nulls, replaces);
 
 		simple_heap_update(pg_rewrite_desc, &tup->t_self, tup);
@@ -137,25 +137,24 @@ InsertRule(char *rulname,
 
 	/* If replacing, get rid of old dependencies and make new ones */
 	if (is_update)
-		deleteDependencyRecordsFor(RelationGetRelid(pg_rewrite_desc),
-								   rewriteObjectId);
+		deleteDependencyRecordsFor(RewriteRelationId, rewriteObjectId);
 
 	/*
 	 * Install dependency on rule's relation to ensure it will go away on
 	 * relation deletion.  If the rule is ON SELECT, make the dependency
-	 * implicit --- this prevents deleting a view's SELECT rule.  Other
-	 * kinds of rules can be AUTO.
+	 * implicit --- this prevents deleting a view's SELECT rule.  Other kinds
+	 * of rules can be AUTO.
 	 */
-	myself.classId = RelationGetRelid(pg_rewrite_desc);
+	myself.classId = RewriteRelationId;
 	myself.objectId = rewriteObjectId;
 	myself.objectSubId = 0;
 
-	referenced.classId = RelOid_pg_class;
+	referenced.classId = RelationRelationId;
 	referenced.objectId = eventrel_oid;
 	referenced.objectSubId = 0;
 
 	recordDependencyOn(&myself, &referenced,
-		 (evtype == CMD_SELECT) ? DEPENDENCY_INTERNAL : DEPENDENCY_AUTO);
+			 (evtype == CMD_SELECT) ? DEPENDENCY_INTERNAL : DEPENDENCY_AUTO);
 
 	/*
 	 * Also install dependencies on objects referenced in action and qual.
@@ -166,7 +165,7 @@ InsertRule(char *rulname,
 	if (event_qual != NULL)
 	{
 		/* Find query containing OLD/NEW rtable entries */
-		Query	   *qry = (Query *) lfirst(action);
+		Query	   *qry = (Query *) linitial(action);
 
 		qry = getInsertSelectQuery(qry, NULL);
 		recordDependencyOnExpr(&myself, event_qual, qry->rtable,
@@ -191,19 +190,16 @@ DefineQueryRewrite(RuleStmt *stmt)
 	Oid			ev_relid;
 	Oid			ruleId;
 	int			event_attno;
-	Oid			event_attype;
-	List	   *l;
+	ListCell   *l;
 	Query	   *query;
-	AclResult	aclresult;
 	bool		RelisBecomingView = false;
 
 	/*
 	 * If we are installing an ON SELECT rule, we had better grab
-	 * AccessExclusiveLock to ensure no SELECTs are currently running on
-	 * the event relation.	For other types of rules, it might be
-	 * sufficient to grab ShareLock to lock out insert/update/delete
-	 * actions.  But for now, let's just grab AccessExclusiveLock all the
-	 * time.
+	 * AccessExclusiveLock to ensure no SELECTs are currently running on the
+	 * event relation.	For other types of rules, it might be sufficient to
+	 * grab ShareLock to lock out insert/update/delete actions.  But for now,
+	 * let's just grab AccessExclusiveLock all the time.
 	 */
 	event_relation = heap_openrv(event_obj, AccessExclusiveLock);
 	ev_relid = RelationGetRelid(event_relation);
@@ -211,9 +207,8 @@ DefineQueryRewrite(RuleStmt *stmt)
 	/*
 	 * Check user has permission to apply rules to this relation.
 	 */
-	aclresult = pg_class_aclcheck(ev_relid, GetUserId(), ACL_RULE);
-	if (aclresult != ACLCHECK_OK)
-		aclcheck_error(aclresult, ACL_KIND_CLASS,
+	if (!pg_class_ownercheck(ev_relid, GetUserId()))
+		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_CLASS,
 					   RelationGetRelationName(event_relation));
 
 	/*
@@ -239,27 +234,23 @@ DefineQueryRewrite(RuleStmt *stmt)
 					 errhint("Use triggers instead.")));
 	}
 
-	/*
-	 * Rules ON SELECT are restricted to view definitions
-	 */
 	if (event_type == CMD_SELECT)
 	{
-		List	   *tllist;
-		int			i;
-
 		/*
+		 * Rules ON SELECT are restricted to view definitions
+		 *
 		 * So there cannot be INSTEAD NOTHING, ...
 		 */
-		if (length(action) == 0)
+		if (list_length(action) == 0)
 			ereport(ERROR,
 					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-			errmsg("INSTEAD NOTHING rules on SELECT are not implemented"),
+			   errmsg("INSTEAD NOTHING rules on SELECT are not implemented"),
 					 errhint("Use views instead.")));
 
 		/*
 		 * ... there cannot be multiple actions, ...
 		 */
-		if (length(action) > 1)
+		if (list_length(action) > 1)
 			ereport(ERROR,
 					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 					 errmsg("multiple actions for rules on SELECT are not implemented")));
@@ -267,11 +258,12 @@ DefineQueryRewrite(RuleStmt *stmt)
 		/*
 		 * ... the one action must be a SELECT, ...
 		 */
-		query = (Query *) lfirst(action);
-		if (!is_instead || query->commandType != CMD_SELECT)
+		query = (Query *) linitial(action);
+		if (!is_instead ||
+			query->commandType != CMD_SELECT || query->into != NULL)
 			ereport(ERROR,
 					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-					 errmsg("rules on SELECT must have action INSTEAD SELECT")));
+				 errmsg("rules on SELECT must have action INSTEAD SELECT")));
 
 		/*
 		 * ... there can be no rule qual, ...
@@ -285,72 +277,17 @@ DefineQueryRewrite(RuleStmt *stmt)
 		 * ... the targetlist of the SELECT action must exactly match the
 		 * event relation, ...
 		 */
-		i = 0;
-		foreach(tllist, query->targetList)
-		{
-			TargetEntry *tle = (TargetEntry *) lfirst(tllist);
-			Resdom	   *resdom = tle->resdom;
-			Form_pg_attribute attr;
-			char	   *attname;
-
-			if (resdom->resjunk)
-				continue;
-			i++;
-			if (i > event_relation->rd_att->natts)
-				ereport(ERROR,
-						(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
-						 errmsg("SELECT rule's target list has too many entries")));
-
-			attr = event_relation->rd_att->attrs[i - 1];
-			attname = NameStr(attr->attname);
-
-			/*
-			 * Disallow dropped columns in the relation.  This won't
-			 * happen in the cases we actually care about (namely creating
-			 * a view via CREATE TABLE then CREATE RULE).  Trying to cope
-			 * with it is much more trouble than it's worth, because we'd
-			 * have to modify the rule to insert dummy NULLs at the right
-			 * positions.
-			 */
-			if (attr->attisdropped)
-				ereport(ERROR,
-						(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-						 errmsg("cannot convert relation containing dropped columns to view")));
-
-			if (strcmp(resdom->resname, attname) != 0)
-				ereport(ERROR,
-						(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
-						 errmsg("SELECT rule's target entry %d has different column name from \"%s\"", i, attname)));
-
-			if (attr->atttypid != resdom->restype)
-				ereport(ERROR,
-						(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
-						 errmsg("SELECT rule's target entry %d has different type from column \"%s\"", i, attname)));
-
-			/*
-			 * Allow typmods to be different only if one of them is -1,
-			 * ie, "unspecified".  This is necessary for cases like
-			 * "numeric", where the table will have a filled-in default
-			 * length but the select rule's expression will probably have
-			 * typmod = -1.
-			 */
-			if (attr->atttypmod != resdom->restypmod &&
-				attr->atttypmod != -1 && resdom->restypmod != -1)
-				ereport(ERROR,
-						(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
-						 errmsg("SELECT rule's target entry %d has different size from column \"%s\"", i, attname)));
-		}
-
-		if (i != event_relation->rd_att->natts)
-			ereport(ERROR,
-					(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
-			   errmsg("SELECT rule's target list has too few entries")));
+		checkRuleResultList(query->targetList,
+							RelationGetDescr(event_relation),
+							true);
 
 		/*
 		 * ... there must not be another ON SELECT rule already ...
 		 */
 		if (!replace && event_relation->rd_rules != NULL)
 		{
+			int			i;
+
 			for (i = 0; i < event_relation->rd_rules->numLocks; i++)
 			{
 				RewriteRule *rule;
@@ -358,9 +295,9 @@ DefineQueryRewrite(RuleStmt *stmt)
 				rule = event_relation->rd_rules->rules[i];
 				if (rule->event == CMD_SELECT)
 					ereport(ERROR,
-					  (errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
-					   errmsg("\"%s\" is already a view",
-							  RelationGetRelationName(event_relation))));
+						  (errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
+						   errmsg("\"%s\" is already a view",
+								  RelationGetRelationName(event_relation))));
 			}
 		}
 
@@ -370,30 +307,30 @@ DefineQueryRewrite(RuleStmt *stmt)
 		if (strcmp(stmt->rulename, ViewSelectRuleName) != 0)
 		{
 			/*
-			 * In versions before 7.3, the expected name was _RETviewname.
-			 * For backwards compatibility with old pg_dump output, accept
-			 * that and silently change it to _RETURN.	Since this is just
-			 * a quick backwards-compatibility hack, limit the number of
-			 * characters checked to a few less than NAMEDATALEN; this
-			 * saves having to worry about where a multibyte character
-			 * might have gotten truncated.
+			 * In versions before 7.3, the expected name was _RETviewname. For
+			 * backwards compatibility with old pg_dump output, accept that
+			 * and silently change it to _RETURN.  Since this is just a quick
+			 * backwards-compatibility hack, limit the number of characters
+			 * checked to a few less than NAMEDATALEN; this saves having to
+			 * worry about where a multibyte character might have gotten
+			 * truncated.
 			 */
 			if (strncmp(stmt->rulename, "_RET", 4) != 0 ||
 				strncmp(stmt->rulename + 4, event_obj->relname,
 						NAMEDATALEN - 4 - 4) != 0)
 				ereport(ERROR,
 						(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
-					  errmsg("view rule for \"%s\" must be named \"%s\"",
-							 event_obj->relname, ViewSelectRuleName)));
+						 errmsg("view rule for \"%s\" must be named \"%s\"",
+								event_obj->relname, ViewSelectRuleName)));
 			stmt->rulename = pstrdup(ViewSelectRuleName);
 		}
 
 		/*
 		 * Are we converting a relation to a view?
 		 *
-		 * If so, check that the relation is empty because the storage for
-		 * the relation is going to be deleted.  Also insist that the rel
-		 * not have any triggers, indexes, or child tables.
+		 * If so, check that the relation is empty because the storage for the
+		 * relation is going to be deleted.  Also insist that the rel not have
+		 * any triggers, indexes, or child tables.
 		 */
 		if (event_relation->rd_rel->relkind != RELKIND_VIEW)
 		{
@@ -402,9 +339,9 @@ DefineQueryRewrite(RuleStmt *stmt)
 			scanDesc = heap_beginscan(event_relation, SnapshotNow, 0, NULL);
 			if (heap_getnext(scanDesc, ForwardScanDirection) != NULL)
 				ereport(ERROR,
-					  (errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
-				errmsg("could not convert table \"%s\" to a view because it is not empty",
-					   event_obj->relname)));
+						(errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
+						 errmsg("could not convert table \"%s\" to a view because it is not empty",
+								event_obj->relname)));
 			heap_endscan(scanDesc);
 
 			if (event_relation->rd_rel->reltriggers != 0)
@@ -429,24 +366,47 @@ DefineQueryRewrite(RuleStmt *stmt)
 			RelisBecomingView = true;
 		}
 	}
+	else
+	{
+		/*
+		 * For non-SELECT rules, a RETURNING list can appear in at most one of
+		 * the actions ... and there can't be any RETURNING list at all in a
+		 * conditional or non-INSTEAD rule.  (Actually, there can be at most
+		 * one RETURNING list across all rules on the same event, but it seems
+		 * best to enforce that at rule expansion time.)  If there is a
+		 * RETURNING list, it must match the event relation.
+		 */
+		bool		haveReturning = false;
+
+		foreach(l, action)
+		{
+			query = (Query *) lfirst(l);
+
+			if (!query->returningList)
+				continue;
+			if (haveReturning)
+				ereport(ERROR,
+						(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+				  errmsg("cannot have multiple RETURNING lists in a rule")));
+			haveReturning = true;
+			if (event_qual != NULL)
+				ereport(ERROR,
+						(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+						 errmsg("RETURNING lists are not supported in conditional rules")));
+			if (!is_instead)
+				ereport(ERROR,
+						(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+						 errmsg("RETURNING lists are not supported in non-INSTEAD rules")));
+			checkRuleResultList(query->returningList,
+								RelationGetDescr(event_relation),
+								false);
+		}
+	}
 
 	/*
 	 * This rule is allowed - prepare to install it.
 	 */
 	event_attno = -1;
-	event_attype = InvalidOid;
-
-	/*
-	 * We want the rule's table references to be checked as though by the
-	 * rule owner, not the user referencing the rule.  Therefore, scan
-	 * through the rule's rtables and set the checkAsUser field on all
-	 * rtable entries.
-	 */
-	foreach(l, action)
-	{
-		query = (Query *) lfirst(l);
-		setRuleCheckAsUser(query, GetUserId());
-	}
 
 	/* discard rule if it's null action and not INSTEAD; it's a no-op */
 	if (action != NIL || is_instead)
@@ -466,42 +426,153 @@ DefineQueryRewrite(RuleStmt *stmt)
 		 * relation is now a view.
 		 *
 		 * Important side effect: an SI notice is broadcast to force all
-		 * backends (including me!) to update relcache entries with the
-		 * new rule.
+		 * backends (including me!) to update relcache entries with the new
+		 * rule.
 		 */
 		SetRelationRuleStatus(ev_relid, true, RelisBecomingView);
 	}
 
 	/*
-	 * IF the relation is becoming a view, delete the storage files
-	 * associated with it.	NB: we had better have AccessExclusiveLock to
-	 * do this ...
+	 * IF the relation is becoming a view, delete the storage files associated
+	 * with it.  NB: we had better have AccessExclusiveLock to do this ...
 	 *
 	 * XXX what about getting rid of its TOAST table?  For now, we don't.
 	 */
 	if (RelisBecomingView)
-		smgrunlink(DEFAULT_SMGR, event_relation);
+	{
+		RelationOpenSmgr(event_relation);
+		smgrscheduleunlink(event_relation->rd_smgr, event_relation->rd_istemp);
+	}
 
 	/* Close rel, but keep lock till commit... */
 	heap_close(event_relation, NoLock);
 }
 
+/*
+ * checkRuleResultList
+ *		Verify that targetList produces output compatible with a tupledesc
+ *
+ * The targetList might be either a SELECT targetlist, or a RETURNING list;
+ * isSelect tells which.  (This is mostly used for choosing error messages,
+ * but also we don't enforce column name matching for RETURNING.)
+ */
+static void
+checkRuleResultList(List *targetList, TupleDesc resultDesc, bool isSelect)
+{
+	ListCell   *tllist;
+	int			i;
+
+	i = 0;
+	foreach(tllist, targetList)
+	{
+		TargetEntry *tle = (TargetEntry *) lfirst(tllist);
+		int32		tletypmod;
+		Form_pg_attribute attr;
+		char	   *attname;
+
+		/* resjunk entries may be ignored */
+		if (tle->resjunk)
+			continue;
+		i++;
+		if (i > resultDesc->natts)
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
+					 isSelect ?
+				   errmsg("SELECT rule's target list has too many entries") :
+					 errmsg("RETURNING list has too many entries")));
+
+		attr = resultDesc->attrs[i - 1];
+		attname = NameStr(attr->attname);
+
+		/*
+		 * Disallow dropped columns in the relation.  This won't happen in the
+		 * cases we actually care about (namely creating a view via CREATE
+		 * TABLE then CREATE RULE, or adding a RETURNING rule to a view).
+		 * Trying to cope with it is much more trouble than it's worth,
+		 * because we'd have to modify the rule to insert dummy NULLs at the
+		 * right positions.
+		 */
+		if (attr->attisdropped)
+			ereport(ERROR,
+					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+					 errmsg("cannot convert relation containing dropped columns to view")));
+
+		if (isSelect && strcmp(tle->resname, attname) != 0)
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
+					 errmsg("SELECT rule's target entry %d has different column name from \"%s\"", i, attname)));
+
+		if (attr->atttypid != exprType((Node *) tle->expr))
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
+					 isSelect ?
+					 errmsg("SELECT rule's target entry %d has different type from column \"%s\"",
+							i, attname) :
+					 errmsg("RETURNING list's entry %d has different type from column \"%s\"",
+							i, attname)));
+
+		/*
+		 * Allow typmods to be different only if one of them is -1, ie,
+		 * "unspecified".  This is necessary for cases like "numeric", where
+		 * the table will have a filled-in default length but the select
+		 * rule's expression will probably have typmod = -1.
+		 */
+		tletypmod = exprTypmod((Node *) tle->expr);
+		if (attr->atttypmod != tletypmod &&
+			attr->atttypmod != -1 && tletypmod != -1)
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
+					 isSelect ?
+					 errmsg("SELECT rule's target entry %d has different size from column \"%s\"",
+							i, attname) :
+					 errmsg("RETURNING list's entry %d has different size from column \"%s\"",
+							i, attname)));
+	}
+
+	if (i != resultDesc->natts)
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
+				 isSelect ?
+				 errmsg("SELECT rule's target list has too few entries") :
+				 errmsg("RETURNING list has too few entries")));
+}
+
 /*
  * setRuleCheckAsUser
- *		Recursively scan a query and set the checkAsUser field to the
- *		given userid in all rtable entries.
+ *		Recursively scan a query or expression tree and set the checkAsUser
+ *		field to the given userid in all rtable entries.
  *
  * Note: for a view (ON SELECT rule), the checkAsUser field of the *OLD*
  * RTE entry will be overridden when the view rule is expanded, and the
  * checkAsUser field of the *NEW* entry is irrelevant because that entry's
- * checkFor bits will never be set.  However, for other types of rules it's
- * important to set these fields to match the rule owner.  So we just set
+ * requiredPerms bits will always be zero.	However, for other types of rules
+ * it's important to set these fields to match the rule owner.  So we just set
  * them always.
  */
-static void
-setRuleCheckAsUser(Query *qry, AclId userid)
+void
+setRuleCheckAsUser(Node *node, Oid userid)
 {
-	List	   *l;
+	(void) setRuleCheckAsUser_walker(node, &userid);
+}
+
+static bool
+setRuleCheckAsUser_walker(Node *node, Oid *context)
+{
+	if (node == NULL)
+		return false;
+	if (IsA(node, Query))
+	{
+		setRuleCheckAsUser_Query((Query *) node, *context);
+		return false;
+	}
+	return expression_tree_walker(node, setRuleCheckAsUser_walker,
+								  (void *) context);
+}
+
+static void
+setRuleCheckAsUser_Query(Query *qry, Oid userid)
+{
+	ListCell   *l;
 
 	/* Set all the RTEs in this query node */
 	foreach(l, qry->rtable)
@@ -511,7 +582,7 @@ setRuleCheckAsUser(Query *qry, AclId userid)
 		if (rte->rtekind == RTE_SUBQUERY)
 		{
 			/* Recurse into subquery in FROM */
-			setRuleCheckAsUser(rte->subquery, userid);
+			setRuleCheckAsUser_Query(rte->subquery, userid);
 		}
 		else
 			rte->checkAsUser = userid;
@@ -524,31 +595,13 @@ setRuleCheckAsUser(Query *qry, AclId userid)
 						  QTW_IGNORE_RT_SUBQUERIES);
 }
 
-/*
- * Expression-tree walker to find sublink queries
- */
-static bool
-setRuleCheckAsUser_walker(Node *node, Oid *context)
-{
-	if (node == NULL)
-		return false;
-	if (IsA(node, Query))
-	{
-		Query	   *qry = (Query *) node;
-
-		setRuleCheckAsUser(qry, *context);
-		return false;
-	}
-	return expression_tree_walker(node, setRuleCheckAsUser_walker,
-								  (void *) context);
-}
-
 
 /*
  * Rename an existing rewrite rule.
  *
  * This is unused code at the moment.
  */
+#ifdef NOT_USED
 void
 RenameRewriteRule(Oid owningRel, const char *oldName,
 				  const char *newName)
@@ -556,7 +609,7 @@ RenameRewriteRule(Oid owningRel, const char *oldName,
 	Relation	pg_rewrite_desc;
 	HeapTuple	ruletup;
 
-	pg_rewrite_desc = heap_openr(RewriteRelationName, RowExclusiveLock);
+	pg_rewrite_desc = heap_open(RewriteRelationId, RowExclusiveLock);
 
 	ruletup = SearchSysCacheCopy(RULERELNAME,
 								 ObjectIdGetDatum(owningRel),
@@ -585,3 +638,5 @@ RenameRewriteRule(Oid owningRel, const char *oldName,
 	heap_freetuple(ruletup);
 	heap_close(pg_rewrite_desc, RowExclusiveLock);
 }
+
+#endif
