@@ -4,56 +4,75 @@
  *	  prototypes for functions in backend/catalog/heap.c
  *
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2006, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
- * $Id: heap.h,v 1.62 2003/09/19 21:04:20 tgl Exp $
+ * $PostgreSQL: pgsql/src/include/catalog/heap.h,v 1.85 2006/07/13 16:49:19 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
 #ifndef HEAP_H
 #define HEAP_H
 
-#include "catalog/pg_attribute.h"
-#include "nodes/parsenodes.h"
 #include "parser/parse_node.h"
-#include "utils/rel.h"
 
 
 typedef struct RawColumnDefault
 {
 	AttrNumber	attnum;			/* attribute to attach default to */
-	Node	   *raw_default;	/* default value (untransformed parse
-								 * tree) */
+	Node	   *raw_default;	/* default value (untransformed parse tree) */
 } RawColumnDefault;
 
+typedef struct CookedConstraint
+{
+	ConstrType	contype;		/* CONSTR_DEFAULT or CONSTR_CHECK */
+	char	   *name;			/* name, or NULL if none */
+	AttrNumber	attnum;			/* which attr (only for DEFAULT) */
+	Node	   *expr;			/* transformed default or check expr */
+} CookedConstraint;
+
 extern Relation heap_create(const char *relname,
 			Oid relnamespace,
+			Oid reltablespace,
+			Oid relid,
 			TupleDesc tupDesc,
+			char relkind,
 			bool shared_relation,
-			bool storage_create,
 			bool allow_system_table_mods);
 
-extern void heap_storage_create(Relation rel);
-
 extern Oid heap_create_with_catalog(const char *relname,
 						 Oid relnamespace,
+						 Oid reltablespace,
+						 Oid relid,
+						 Oid ownerid,
 						 TupleDesc tupdesc,
 						 char relkind,
 						 bool shared_relation,
+						 bool oidislocal,
+						 int oidinhcount,
 						 OnCommitAction oncommit,
+						 Datum reloptions,
 						 bool allow_system_table_mods);
 
-extern void heap_drop_with_catalog(Oid rid);
+extern void heap_drop_with_catalog(Oid relid);
 
-extern void heap_truncate(Oid rid);
+extern void heap_truncate(List *relids);
 
-extern void heap_truncate_check_FKs(Relation rel);
+extern void heap_truncate_check_FKs(List *relations, bool tempTables);
 
-extern void AddRelationRawConstraints(Relation rel,
+extern List *heap_truncate_find_FKs(List *relationIds);
+
+extern void InsertPgClassTuple(Relation pg_class_desc,
+				   Relation new_rel_desc,
+				   Oid new_rel_oid,
+				   Datum reloptions);
+
+extern List *AddRelationRawConstraints(Relation rel,
 						  List *rawColDefaults,
 						  List *rawConstraints);
 
+extern void StoreAttrDefault(Relation rel, AttrNumber attnum, char *adbin);
+
 extern Node *cookDefault(ParseState *pstate,
 			Node *raw_default,
 			Oid atttypid,
@@ -69,6 +88,7 @@ extern void RemoveAttributeById(Oid relid, AttrNumber attnum);
 extern void RemoveAttrDefault(Oid relid, AttrNumber attnum,
 				  DropBehavior behavior, bool complain);
 extern void RemoveAttrDefaultById(Oid attrdefId);
+extern void RemoveStatistics(Oid relid, AttrNumber attnum);
 
 extern Form_pg_attribute SystemAttributeDefinition(AttrNumber attno,
 						  bool relhasoids);
